# :title: XMLStreamin Documentation
# XMLStreamin is a small module that provides a way of reading XML as a stream,
# while letting it be processed acccording to its tree structure much more
# handily than the usual 'flat' stream reader does.
#
# Unlike such readers, which typically simply call unspecialized methods
# for each start-tag, end-tag, text segment, and so on, and leave it to the
# application to sort out the hierarchy, XMLStreamin uses a pre-built tree
# of #XMLSpec nodes to model the expected document structure.
#
# Each node specifies the actual actions to be taken when an element that
# it represents is encountered.  It can specify what processing needs to be
# done on the attributes of a start-tag, the handling of included text, and any
# clean-up actions when the end-tag is read.  It also contains a table of the
# expected sub-elements and _their_ XMLSpec nodes, thus reflecting the document
# tree.
#
# Usage skeleton:
#	 require "xmlstreamin"
#	 include XMLStreamin # for convenience in naming
# 
#	 # Supply XMLSpec objects to model the document tree: 
# 
#	 #First for the leaf elements:
#	 someleafspec = XMLSpec.new
#	 # override methods as needed, e.g.:
#	 def someleafspec.text(context,data)
#	 	#...
#	 end
# 
#	 #... then provide nodes for the elements that enclose these:
#	 someotherspec = XMLSpec.new
#	 someotherspec.specs!({'someleaftag'=>someleafspec,...})
#	 # Override methods in these as well, if appropriate:
#	 def someotherspec.start(context,name,attrs)
#	 	#...
#	 	return context
#	 end
# 
#	 #...... more specs complete the tree, up to some 'toplevelspec'
#	 # for the document's enclosing element.
# 
#	 # Finally one spec for the document itself that just has
#	 # an entry for that top level tag:
#	 specDocument = XMLSpec.new
#	 specDocument.specs!({'document_top_tag'=>toplevelspec})
# 
#	 To run, provide a source stream, and invoke the REXML stream parser:
#	 source = ...the source of the document stream 
#	 REXML::Document.parse_stream(source,
#	 		XMLStreamListener.new(specDocument))
#
#
module XMLStreamin

 require "rexml/document"
 require "rexml/streamlistener"

	# Exception class that will be thrown by the module if it hits trouble.
	class XMLError < RuntimeError
	end
	
	# XMLSpec is a base class intended to be extended as needed to handle
	# processing for each type of expected element at each level of the
	# XML hierarchy in the document.  It has two categories of methods:
	# those concerned with setup (which should not need to be modified)
	# -- #specs!, #default!, and #spec --, and the handler methods
	# #start, #done, #empty and #text, that are intended to be
	# specialized as necessary in derived classes or instances.
	#
	# XMLSpec nodes are intended to be linked in a tree structure, reflecting
	# the structure of the XML document to be read.  Each node has a 'dispatch'
	# (hash) table associating expected tag names with the subordinate XMLSpec
	# nodes that should handle them; the hash table _default_ should reference
	# a node that will handle unexpected tags.  If appropriate, you can use
	# a single node to service several different tag names (bearing in mind
	# that the dispatch table will be shared).
	#
	# There are two predefined global XMLSpecs:
	# $specXMLVoid::
	#  This is a do-nothing basic XMLSpec that can be used to represent
	#  elements that you aren't interested in.  These and any subordinate
	#  elements will be skipped during parsing.
	# $specXMLFail::
	#  This will raise XMLError if it is invoked.  It may be used if an
	#  unexpected element is a serious problem.
	class XMLSpec
		# Create an empty instance.  Unless it is a non-functional leaf node,
		# it will need to be specialized by filling the dispatch table ( #specs! )
		# and redefining methods as appropriate.
		def initialize
			@subspecs=Hash.new($specXMLVoid) end

		# Set the XMLSpec that will be used for tag names that are not
		# specifically referenced by the dispatch table.
		def default! spec
			@subspecs.default = spec end

		# Adds tagnames and their associated XMLSpecs to the dispatch table.
		# The argument _specs_ is a preconstructed hash that will be merged
		# with any existing table.  (There is no way of directly removing
		# entries -- except by overwriting them -- as the tree is a static
		# structure built before reading the document.)
		def specs! specs
			@subspecs.merge! specs end

		# Convenience method for accessing the dispatch table; returns the hash.	 
		def specs
			return @subspecs end

		# Locates and returns the XMLSpec that handles _name_ in the dispatch table.	 
		def spec name
			return @subspecs[name] end

		# Called by XMLStreamListener when a start-tag for an element handled
		# by this XMLSpec is read. Arguments:
		# _context_::
		#   the context object passed from the level above (application
		#   specific -- may be nil).  This is also the return value
		#   from the method, and will be passed on to any subordinate
		#   nodes.  A derived version may actually supply a completely
		#   new context if desired, but if it does so, it _must_
		#   preserve the one it received (in an instance variable)
		#   for restoration later by the #done (or #empty) method.
		#   (In any normal situation the same context object --
		#   updated as required -- will be used throughout.  No
		#   special preservation is then needed, as long as all
		#   methods return it on exit.)
		# _name_::
		#   the actual tag name of the element.  (The same node can handle
		#   several tagnames.)
		# _attrs_::	a hash of the attribute _name_/_value_ pairs in the tag.
		# The base method is a dummy that does nothing but return the context.
		# Derived nodes will redefine the method as necessary.
		def start context,name,attrs
	 		return context	# can get changed
		end
	 
		# Called by XMLStreamListener when an end-tag for an element handled
		# by this XMLSpec is read _and_ the element was not empty.
		# If it is actually an empty element #empty in invoked instead.
		# If any action needs to be taken at the end of an element, this dummy
		# method may be redefined.
		# Arguments:
		# _context_::
		#   the context object at this level (application
		#   specific -- may be nil).  This is also the return value
		#   from the method, and will be passed back to XMLStreamListener
		#   (which does _not_ keep track of context itself!).
		#   If the #start method provided a new context this _must_
		#   restore the one preserved at that time.
		# _name_::
		#   the actual tag name of the element.  (Not normally needed,
		#   but it might be useful to have the name again here.)
		def done context,name
	 		return context	# can get restored
		end
	 
		# Called by XMLStreamListener when an end-tag for an element handled
		# by this XMLSpec is read _and_ the element was empty.
		# If it is not actually an empty element #done in invoked instead.
		# If any action needs to be taken for an empty element, this dummy
		# method may be redefined.
		# Arguments:
		# _context_::
		#   the context object at this level (application
		#   specific -- may be nil).  This is also the return value
		#   from the method, and will be passed back to XMLStreamListener
		#   (which does _not_ keep track of context itself!).
		#   If the #start method provided a new context this _must_
		#   restore the one preserved at that time.
		# _name_::
		#   the actual tag name of the element.  (Not normally needed,
		#   but it might be useful to have the name again here.)
		def empty context
			# called only if no enclosed text or elements
			return context	
		end
	 
		# Called by XMLStreamListener when text is encountered within a
		# handled element.  It will be invoked separately for each text
		# segment (separated by subordinate elements) within the element.
		# It is a dummy method that may be redefined as desired to handle
		# the text.  It does not need to return the _context_.
		def text context,data
		end

		# Called by XMLStreamListener when CDATA is encountered within a
		# handled element.  
		# It is a dummy method that may be redefined as desired to handle
		# the content.  It does not need to return the _context_.
		def cdata context,content
		end
	end

	# Predefined global XMLSpecs:

	$specXMLVoid = XMLSpec.new
	$specXMLVoid.default!($specXMLVoid)	# to avoid circularity

	$specXMLFail = XMLSpec.new
	$specXMLFail.default!($specXMLFail)
	def $specXMLFail.start(context,name,attrs)
		raise XMLError.new("Failed Tag <#{name}...>")
	end

	# This class extends REXML::StreamListener to provide an interface to
	# a 'tree' of XMLSpec nodes that models the hierarchy of the XML document
	# to be read.
	class XMLStreamListener
	 include REXML::StreamListener
	 	# Create a new XMLStreamListener with _root_ as the root XMLSpec
	 	# of the XML hierarchy to be parsed. _base_ is an optional 'context',
	 	# of any form suitable to the task, that will be passed to all XMLSpec
	 	# methods invoked.
		def initialize root=$specXMLVoid, base=nil
			@currSpec=root
			@currContext=base
			@prevspecs=[]
			@openTag=nil	
		end
		# Invoked when a tag is encountered, with args:
		# * _name_ the tag name
		# * _attrs_ a Hash of attribute/value pairs. [*NOT* an array of arrays!]
		#   -- i.e. a start tag like:
		#	 <tag attr1="value1" attr2="value2">
		#   will result in: 
		#   tag_start( "tag", {"attr1"=>"value1","attr2"=>"value2"})
		#
		# This in turn determines the appropriate XMLSpec node that should
		# handle the tag (by querying the current spec), sets this as the new
		# current spec, and invokes its XMLSpec#start method.
		def tag_start name, attrs
			@prevspecs.push(@currSpec)
			@openTag=name
			@currSpec = @currSpec.spec name
		 	@currContext = @currSpec.start(@currContext,name,attrs)
		end
		# Invoked when the end tag is reached, with the _name_ of the tag
		# as argument.  In the case of  an empty tag ('<tag/>',
		# tag_end will be called immediately after tag_start.
		# If the element was not empty, the current XMLSpec#done method is
		# called, otherwise the XMLSpec#empty method.  Then the previous
		# higher level) spec is restored.
		def tag_end name
			@currContext = if (@openTag == name)
				@currSpec.empty(@currContext)
			  else
				@currSpec.done(@currContext, name)
			  end
			@openTag=nil	
			@currSpec = @prevspecs.pop
		end
		# Invoked when text is encountered in the document,
		# with the _text_ content as argument.
		# The current XMLSpec#text method is in turn called.
		# (Note that if the text is interspersed with other elements,
		# this method is invoked for each segment separately.)
		def text text
			@openTag=nil	
			@currSpec.text(@currContext, text)
		end
		# Invoked when a CDATA tag is encountered in the document,
		# with the _content_ as argument.
		# The current XMLSpec#cdata method is in turn called.
		def cdata content
			@openTag=nil	
			@currSpec.cdata(@currContext, content)
		end
		def entitydecl ent,value
# 			print "ENTITY DECL: #{ent} --> #{value}\n"
		end
		def entity ent	### This is NEVER called!!
		end
	end

end

